module CopAlarmed where

import Control.Concurrent.MVar
import ConcurrentUtil
import State
import AI
import LegalMoves
import RandomUtils
import Data.Maybe
import RobberTracking
import Evaluate
import Minimax
import Data.List
import System.IO
import PlotCourse

-- cop-alarmed moves to wherever the robber was last seen
-- never shares information, never submits plans,
-- and votes for himself followed by a random permutation of cop names

cop_alarmed = CopAI { cai_name = "cop-alarmed",
                     cai_init_type = PT_cop_foot,
                     cai_init = cop_alarmed_init,
                     cai_inform = cop_alarmed_inform,
                     cai_plan = cop_alarmed_plan,
                     cai_vote = cop_alarmed_vote,
                     cai_move = cop_alarmed_move }

data AlarmedAIState = AlarmedAIState {
    aas_ws :: WorldSkeleton,
    aas_update :: WorldUpdate,
    aas_last_robber_loc :: LocName,
    aas_my_plans :: [Plan]
} deriving Show

cop_alarmed_init :: InitAI AlarmedAIState
cop_alarmed_init skel state_var =
    putMVar state_var (AlarmedAIState {
                           aas_ws = skel,
                           aas_last_robber_loc = ws_robber_start skel,
                           aas_update = error "cop_alarmed_init: no update",
                           aas_my_plans = error "cop_alarmed_init: no plan" })

cop_alarmed_inform :: StepAI AlarmedAIState WorldUpdate [Inform]
cop_alarmed_inform state update result_var =
    do let s' = case [l | (_, PlayerState { ps_loc = l, ps_type = PT_robber })
                          <- wu_players update] of
                [r_loc] -> state { aas_update = update,
                                   aas_last_robber_loc = r_loc }
                _       -> state { aas_update = update }
       putMVarF2 result_var (s', [])

cop_alarmed_plan :: StepAI AlarmedAIState [Inform] [Plan]
cop_alarmed_plan state informs result_var =
    do let ws = aas_ws state
           players = wu_players (aas_update state)
           plan_epoch = wu_epoch (aas_update state) + 1
           [my_ps] = [ps | (name, ps) <- players, name == ws_my_name ws]
           stay_put = Plan { pl_source = ws_my_name ws,
                             pl_epoch = plan_epoch,
                             pl_who = ws_my_name ws,
                             pl_move = my_ps }
       putMVarF2 result_var (state, [stay_put])
       let robber_loc = aas_last_robber_loc state
           next_step_or_stay ps =
              case plot_course ws ps robber_loc of
              ([],_) -> error "cop_alarmed_plan: no course"
              ((here:next_step:_),_) -> next_step
              ([here],_) -> here
           plans_solo =
               [Plan { pl_source = ws_my_name ws,
                       pl_epoch = plan_epoch,
                       pl_who = name,
                       pl_move = next_step_or_stay ps }
                | (name, ps@PlayerState { ps_type = pt }) <- players,
                  pt /= PT_robber && name == ws_my_name ws]
           plans_all =
               [Plan { pl_source = ws_my_name ws,
                       pl_epoch = plan_epoch,
                       pl_who = name,
                       pl_move = next_step_or_stay ps }
                | (name, ps@PlayerState { ps_type = pt }) <- players,
                  pt /= PT_robber]
       swapMVarF2_ result_var (state { aas_my_plans = plans_solo }, plans_solo)

cop_alarmed_vote :: StepAI AlarmedAIState [Plan] Vote
cop_alarmed_vote state plans result_var =
    do let moi = ws_my_name (aas_ws state)
           cops_sans_moi = delete moi (ws_cops (aas_ws state))
       rest_vote <- random_list_perm cops_sans_moi
       putMVarF2 result_var (state, moi:rest_vote)

cop_alarmed_move :: StepAI AlarmedAIState Winner PlayerState
cop_alarmed_move state winner result_var =
    do let -- seed with identity move in case time runs out
           ws = aas_ws state
           players = wu_players (aas_update state)
           [stay_put] = [ps | (name, ps) <- players, name == ws_my_name ws]
       putMVarF2 result_var (state, stay_put)
       let plans = aas_my_plans state
           me = ws_my_name (aas_ws state)
           my_move = case [p | p@(Plan { pl_who = who }) <- plans, who == me] of
                     [plan] -> pl_move plan
                     junk -> error ("cop_alarmed_plan: " ++ show junk)
       swapMVarF2_ result_var (state, my_move)

