module CopBanker where

import Control.Concurrent.MVar
import ConcurrentUtil
import State
import AI
import LegalMoves
import RandomUtils
import SearchMove
import Data.Maybe
import RobberTracking
import RobberTrackingImpl
import Evaluate
import Minimax
import Data.List
import System.IO
import PlotCourse
import ListUtil

cop_banker = CopAI { cai_name = "cop-banker",
                     cai_init_type = PT_cop_foot,
                     cai_init = cop_banker_init,
                     cai_inform = cop_banker_inform,
                     cai_plan = cop_banker_plan,
                     cai_vote = cop_banker_vote,
                     cai_move = cop_banker_move }

data BankerAIState = BankerAIState {
    bas_skeleton :: WorldSkeleton,
    bas_current_ps :: PlayerState,
    bas_update :: WorldUpdate,
    bas_planned_ps_list :: [PlayerState],
    bas_my_impossible_list :: ImpossibleList,
    bas_all_impossible_list :: ImpossibleList
}

cop_banker_init :: InitAI BankerAIState
cop_banker_init skel state_var =
    putMVar state_var (BankerAIState {
                           bas_skeleton = skel,
                           bas_current_ps = PlayerState {
                                               ps_type = PT_cop_foot,
                                               ps_loc = ws_hq skel },
                           bas_update = error "No update",
                           bas_planned_ps_list =
                               error "No planned ps list",
                           bas_my_impossible_list = initImpossibleList skel,
                           bas_all_impossible_list = initImpossibleList skel })

impossibleLocsToInforms :: WorldSkeleton -> BankerAIState -> LocSet -> [Inform]
impossibleLocsToInforms ws state impossibleLocs =
    let update = bas_update state in
        [Inform { if_source = ws_my_name ws,
                  if_who = ws_robber ws,
                  if_loc = (ws_loc_id_to_name ws) locId,
                  if_ptype = PT_robber,
                  if_epoch = wu_epoch update,
                  if_certainty = -100 }
         |
         locId <- lsToList impossibleLocs]

cop_banker_inform :: StepAI BankerAIState WorldUpdate [Inform]
cop_banker_inform state update result_var =
    do let skel = bas_skeleton state
           my_name = ws_my_name skel
           myOldImpossibleList = bas_my_impossible_list state
           myCurImpossibleLocs = copCurImpossibleLocs skel update my_name
           myNewImpossibleList =
               nextImpossibleList skel myOldImpossibleList myCurImpossibleLocs
           next_state =
               state { bas_update = update,
                       bas_my_impossible_list = myNewImpossibleList }
       putMVarF2 result_var (next_state, [])
       -- hPutStrLn stderr $ "Cop inform: no information"

       let informs =
               impossibleLocsToInforms
               skel next_state (head myNewImpossibleList)

       swapMVarF2_ result_var (next_state, informs)
       -- hPutStrLn stderr $ "Cop inform: my impossible location"

       -- hPutStrLn stderr $
       --     "Cop inform: my tracking: possible = " ++
       --     show ((lsSize (ws_all_robber_locs skel)) -
       --           lsSize (head myNewImpossibleList)) ++
       --     ", impossible = " ++
       --     show (lsSize (head myNewImpossibleList))

statesToPlans fromCop epoch cops cop_states =
    map (\(cop, state) ->
             Plan { pl_source = fromCop,
                    pl_epoch = epoch + 1,
                    pl_who = cop,
                    pl_move = state }) (zip cops cop_states)

informsToImpossibleLocs :: WorldSkeleton -> BankerAIState -> [Inform] -> LocSet
informsToImpossibleLocs ws state informs =
    let update = bas_update state in
        lsFromList
        [(ws_loc_name_to_id ws) (if_loc inform)
         |
         inform <- informs,
         if_source inform /= ws_my_name ws,
         if_who inform == ws_robber ws,
         (if_loc inform) `elem` (ws_all_robber_loc_names ws),
         if_ptype inform == PT_robber,
         if_epoch inform == wu_epoch update,
         if_certainty inform == -100]

cop_banker_plan :: StepAI BankerAIState [Inform] [Plan]
cop_banker_plan state informs result_var =
    do let skel = bas_skeleton state
           update = bas_update state

           allOldImpossibleList = bas_all_impossible_list state
           allCurImpossibleLocs =
               (informsToImpossibleLocs skel state informs) `lsUnion`
               (head (bas_my_impossible_list state))
           allNewImpossibleList =
               nextImpossibleList
               skel allOldImpossibleList allCurImpossibleLocs
           cops = ws_cops skel
           cop_states =
               map
               (fromJust . ((flip lookup) (wu_players update)))
               cops
           identity_plans =
               statesToPlans (ws_my_name skel) (wu_epoch update)
                             cops cop_states
           identity_state = state { bas_planned_ps_list = cop_states,
                                    bas_all_impossible_list =
                                        allNewImpossibleList }
       putMVarF2 result_var (identity_state, identity_plans)
       -- hPutStrLn stderr $ "Cop plan: identity"

       random_cop_states <-
           mapM
           random_list_elt
           (map (legalMoves skel) cop_states)
       let random_plans =
               statesToPlans (ws_my_name skel) (wu_epoch update)
                             cops random_cop_states
           random_state = identity_state { bas_planned_ps_list =
                                               random_cop_states }
       swapMVarF2_ result_var (random_state, random_plans)
       -- hPutStrLn stderr $ "Cop plan: random"

       -- hPutStrLn stderr $
       --     "Cop inform: all tracking: possible = " ++
       --     show ((lsSize (ws_all_robber_locs skel)) -
       --           lsSize (head allNewImpossibleList)) ++
       --     ", impossible = " ++
       --     show (lsSize (head allNewImpossibleList))

       let impossibleListToUse =
               if (lsSize (head (bas_all_impossible_list random_state)) ==
                   lsSize (ws_all_robber_locs skel))
               then bas_my_impossible_list random_state
               else bas_all_impossible_list random_state
           impossibleLocsToUse = head impossibleListToUse
           possibleLocsToUse =
               (ws_all_robber_locs skel) `lsDifference` impossibleLocsToUse

       let poss_robber_locs :: [LocName]
           poss_robber_locs = (map (ws_loc_id_to_name skel)
                               (lsToList possibleLocsToUse))

           findOverriddenMoves copState =
               let theLegalMoves = legalMoves skel copState
                   adjacentRobberStates =
                       [nextState
                        |
                        nextState <- theLegalMoves,
                        stateToLoc nextState `elem` (ws_banks skel),
                        stateToLoc nextState `elem` poss_robber_locs] ++
                       [nextState
                        |
                        nextState <- theLegalMoves,
                        not (stateToLoc nextState `elem` (ws_banks skel)),
                        stateToLoc nextState `elem` poss_robber_locs]
                   moveOffBankStates =
                       if stateToLoc copState `elem` (ws_banks skel)
                       then [move |
                             move <- theLegalMoves,
                             move /= copState]
                       else []
               in adjacentRobberStates ++ moveOffBankStates
           overriddenMoves = map findOverriddenMoves cop_states


       -- hPutStrLn stderr $ "Possible robber locs:"
       -- hPutStrLn stderr $ unlines $ map show poss_robber_locs
       -- hPutStrLn stderr $ "Possible robber locs end"
       
       let loc_to_rob_ps l = PlayerState { ps_loc = l, ps_type = PT_robber }
           -- 1. find five banks closest to the robber, on the average
           avg_bank_dist bank_loc =
               let avg [] = 10000 :: Double
                   avg xs = fromInteger (toInteger (sum xs)) /
                            fromInteger (toInteger (length xs))
               in  avg [snd (plot_course skel (loc_to_rob_ps l) bank_loc)
                        | l <- poss_robber_locs]
           closest_banks = map snd $ take (length cops) $ sort $
                           [(avg_bank_dist bank, bank) | bank <- ws_banks skel]
           -- 2. optimally assign cops to banks (considering total distance)
           plot_course_safe skel ps loc =
                 case plot_course skel ps loc of
                   ([],dist) -> error "cop_alarmed_plan: no course"
                   ((here:next_step:_),dist) -> (dist, next_step)
                   ([here],dist) -> (dist, here)
           cost_assign ass =
               let evals = [plot_course_safe skel cop_ps bank_loc | (cop_ps,bank_loc) <- ass]
                   total_dist = sum (map fst evals)
                   moves_per_cop = map snd evals
               in  (total_dist, moves_per_cop)
           assigns = [zip cop_states banks | banks <- permutations closest_banks]
           next_pss = case sort (map cost_assign assigns) of
                      [] -> []
                      ((_, moves_per_cop) : _) -> moves_per_cop

           override_ps (next_ps, override) =
               case override
               of
               [] -> next_ps
               (first_override:_) -> first_override

           overridden_pss = map override_ps (zip next_pss overriddenMoves)

           real_plans = statesToPlans (ws_my_name skel) (wu_epoch update)
                                      cops overridden_pss
           next_state = random_state { bas_planned_ps_list = overridden_pss }
       swapMVarF2_ result_var (next_state, real_plans)
--       hPutStrLn stderr $ "Cop plan: depth " ++ show depth ++ " done"
       -- hPutStrLn stderr $ "Cop plan is: " ++ show next_pss
       --hPutStrLn stderr $ "Cop plan all is: " ++ show nextAll

cop_banker_vote :: StepAI BankerAIState [Plan] Vote
cop_banker_vote state plans result_var =
    do let ws = bas_skeleton state
           my_name = ws_my_name ws
       vote <- random_list_perm (ws_cops ws)
       let vote' = [my_name] ++
                   [name |
                    name <- vote,
                    name /= my_name]
       putMVarF2 result_var (state, vote')

cop_banker_move :: StepAI BankerAIState Winner PlayerState
cop_banker_move state winner result_var =
    do -- seed with identity move in case time runs out
       putMVarF2 result_var (state, bas_current_ps state)
       -- hPutStrLn stderr $ "Cop move: identity"
       let skel = bas_skeleton state
           update = bas_update state
           planned_ps_list = bas_planned_ps_list state
           my_name = ws_my_name skel
           cops = ws_cops skel
           Just idx = elemIndex my_name cops
           next_ps = planned_ps_list !! idx
           next_state = state { bas_current_ps = next_ps }
       swapMVarF2_ result_var (next_state, next_ps)
       -- hPutStrLn stderr $ "Cop move: planned"
