module CopSearch where

import Control.Concurrent.MVar
import ConcurrentUtil
import State
import AI
import LegalMoves
import RandomUtils
import SearchState
import SearchMove
import Data.Maybe
import RobberTracking
import RobberTrackingImpl
import Evaluate
import Minimax
import Data.List
import System.IO
import PlotCourse

-- cop-search moves according to a game tree search,
-- never shares information, never submits
-- plans, and votes a random permutation of cop names

cop_search = CopAI { cai_name = "cop-search",
                     cai_init_type = PT_cop_foot,
                     cai_init = cop_search_init,
                     cai_inform = cop_search_inform,
                     cai_plan = cop_search_plan,
                     cai_vote = cop_search_vote,
                     cai_move = cop_search_move }

data SearchAIState = SearchAIState {
    sas_skeleton :: WorldSkeleton,
    sas_current_ps :: PlayerState,
    sas_update :: WorldUpdate,
    sas_planned_ps_list :: [PlayerState],
    sas_my_impossible_list :: ImpossibleList,
    sas_all_impossible_list :: ImpossibleList
}

cop_search_init :: InitAI SearchAIState
cop_search_init skel state_var =
    putMVar state_var (SearchAIState {
                           sas_skeleton = skel,
                           sas_current_ps = PlayerState {
                                               ps_type = PT_cop_foot,
                                               ps_loc = ws_hq skel },
                           sas_update = error "No update",
                           sas_planned_ps_list =
                               error "No planned ps list",
                           sas_my_impossible_list = initImpossibleList skel,
                           sas_all_impossible_list = initImpossibleList skel })

impossibleLocsToInforms :: WorldSkeleton -> SearchAIState -> LocSet -> [Inform]
impossibleLocsToInforms ws state impossibleLocs =
    let update = sas_update state in
        [Inform { if_source = ws_my_name ws,
                  if_who = ws_robber ws,
                  if_loc = (ws_loc_id_to_name ws) locId,
                  if_ptype = PT_robber,
                  if_epoch = wu_epoch update,
                  if_certainty = -100 }
         |
         locId <- lsToList impossibleLocs]

cop_search_inform :: StepAI SearchAIState WorldUpdate [Inform]
cop_search_inform state update result_var =
    do let skel = sas_skeleton state
           my_name = ws_my_name skel
           myOldImpossibleList = sas_my_impossible_list state
           myCurImpossibleLocs = copCurImpossibleLocs skel update my_name
           myNewImpossibleList =
               nextImpossibleList skel myOldImpossibleList myCurImpossibleLocs
           next_state =
               state { sas_update = update,
                       sas_my_impossible_list = myNewImpossibleList }
       putMVarF2 result_var (next_state, [])
       hPutStrLn stderr $ "Cop inform: no information"

       let informs =
               impossibleLocsToInforms
               skel next_state (head myNewImpossibleList)

       swapMVarF2_ result_var (next_state, informs)
       hPutStrLn stderr $ "Cop inform: my impossible location"

       hPutStrLn stderr $
           "Cop inform: my tracking: possible = " ++
           show ((lsSize (ws_all_robber_locs skel)) -
                 lsSize (head myNewImpossibleList)) ++
           ", impossible = " ++
           show (lsSize (head myNewImpossibleList))

statesToPlans fromCop epoch cops cop_states =
    map (\(cop, state) ->
             Plan { pl_source = fromCop,
                    pl_epoch = epoch + 1,
                    pl_who = cop,
                    pl_move = state }) (zip cops cop_states)

informsToImpossibleLocs :: WorldSkeleton -> SearchAIState -> [Inform] -> LocSet
informsToImpossibleLocs ws state informs =
    let update = sas_update state in
        lsFromList
        [(ws_loc_name_to_id ws) (if_loc inform)
         |
         inform <- informs,
         if_source inform /= ws_my_name ws,
         if_who inform == ws_robber ws,
         (if_loc inform) `elem` (ws_all_robber_loc_names ws),
         if_ptype inform == PT_robber,
         if_epoch inform == wu_epoch update,
         if_certainty inform == -100]

cop_search_plan :: StepAI SearchAIState [Inform] [Plan]
cop_search_plan state informs result_var =
    do let skel = sas_skeleton state
           update = sas_update state

           allOldImpossibleList = sas_all_impossible_list state
           allCurImpossibleLocs =
               (informsToImpossibleLocs skel state informs) `lsUnion`
               (head (sas_my_impossible_list state))
           allNewImpossibleList =
               nextImpossibleList
               skel allOldImpossibleList allCurImpossibleLocs

           cops = ws_cops skel
           cop_states =
               map
               (fromJust . ((flip lookup) (wu_players update)))
               cops
           identity_plans =
               statesToPlans (ws_my_name skel) (wu_epoch update)
                             cops cop_states
           identity_state = state { sas_planned_ps_list = cop_states,
                                    sas_all_impossible_list =
                                        allNewImpossibleList }
       putMVarF2 result_var (identity_state, identity_plans)
       hPutStrLn stderr $ "Cop plan: identity"

       random_cop_states <-
           mapM
           random_list_elt
           (map (legalMoves skel) cop_states)
       let random_plans =
               statesToPlans (ws_my_name skel) (wu_epoch update)
                             cops random_cop_states
           random_state = identity_state { sas_planned_ps_list =
                                               random_cop_states }
       swapMVarF2_ result_var (random_state, random_plans)
       hPutStrLn stderr $ "Cop plan: random"

       hPutStrLn stderr $
           "Cop inform: all tracking: possible = " ++
           show ((lsSize (ws_all_robber_locs skel)) -
                 lsSize (head allNewImpossibleList)) ++
           ", impossible = " ++
           show (lsSize (head allNewImpossibleList))

       let impossibleListToUse =
               if (lsSize (head (sas_all_impossible_list random_state)) ==
                   lsSize (ws_all_robber_locs skel))
               then sas_my_impossible_list random_state
               else sas_all_impossible_list random_state
           impossibleLocsToUse = head impossibleListToUse
           possibleLocsToUse =
               (ws_all_robber_locs skel) `lsDifference` impossibleLocsToUse

       let all_robber_locs = (map (ws_loc_id_to_name skel)
                              (lsToList possibleLocsToUse))
       --hPutStrLn stderr $ "All robber locs:"
       --hPutStrLn stderr $ unlines $ map show all_robber_locs
       --hPutStrLn stderr $ "All robber locs end"
       let next_step_or_stay ps =
              do rand_loc <- random_list_elt all_robber_locs
                 case plot_course skel ps rand_loc of
                   ([],_) -> error "cop_alarmed_plan: no course"
                   ((here:next_step:_),_) -> return next_step
                   ([here],_) -> return here
       next_ps_list <- mapM next_step_or_stay cop_states
       let real_plans =
               statesToPlans (ws_my_name skel) (wu_epoch update)
                             (ws_cops skel) next_ps_list
           next_state = random_state { sas_planned_ps_list = next_ps_list }
       swapMVarF2_ result_var (next_state, real_plans)
--       hPutStrLn stderr $ "Cop plan: depth " ++ show depth ++ " done"
       hPutStrLn stderr $ "Cop plan is: " ++ show next_ps_list
       --hPutStrLn stderr $ "Cop plan all is: " ++ show nextAll
--       cop_search_plan_iterate state result_var ss depth
{-
       hPutStrLn stderr $
           "Cop inform: chosen tracking: possible = " ++
           show (lsSize possibleLocsToUse) ++
           ", impossible = " ++
           show (lsSize impossibleLocsToUse)

       randomRobberLoc <- random_list_elt
                          (map (ws_loc_id_to_name skel)
                           (lsToList possibleLocsToUse))
       let ssFake = SearchState { ss_epoch = (wu_epoch update),
                                  ss_cop_states = cop_states,
                                  ss_robber_state = error "No robber state",
                                  ss_fake_robber_states =
                                      Just ([locToState
                                             PT_robber
                                             randomRobberLoc]),
                                  ss_impossible_list = 
                                      impossibleListToUse }
           fakeDepth = nextDepth skel ssFake 0
       hPutStrLn stderr $ "Cop plan: fake depth " ++ show fakeDepth ++ " start"
       let (fakeNextScore, (_:fakeNextMove@(CopMove fake_next_ps_list):_)) =
               minimax fakeDepth False skel ssFake
           fake_real_plans =
               statesToPlans (ws_my_name skel) (wu_epoch update)
                             (ws_cops skel) fake_next_ps_list
           fake_state = random_state { sas_planned_ps_list = fake_next_ps_list }
       swapMVarF2_ result_var (fake_state, fake_real_plans)
       hPutStrLn stderr $ "Cop plan: fake depth " ++ show fakeDepth ++ " done"
       hPutStrLn stderr $ "Cop plan is: " ++ show (fakeNextScore, fakeNextMove)

       let ss = SearchState { ss_epoch = (wu_epoch update),
                              ss_cop_states = cop_states,
                              ss_robber_state = error "No robber state",
                              ss_fake_robber_states =
                                  Just (map ((locToState PT_robber) .
                                             (ws_loc_id_to_name skel))
                                            (lsToList possibleLocsToUse)),
                              ss_impossible_list = 
                                  impossibleListToUse }
       cop_search_plan_iterate random_state result_var ss 0
-}

cop_search_plan_iterate state result_var ss prevDepth =
    do let skel = sas_skeleton state
           update = sas_update state
           depth = nextDepth skel ss prevDepth
       hPutStrLn stderr $ "Cop plan: depth " ++ show depth ++ " start"
       let nextAll@(nextScore, (_:nextMove@(CopMove next_ps_list):_)) =
               minimax depth False skel ss
           real_plans =
               statesToPlans (ws_my_name skel) (wu_epoch update)
                             (ws_cops skel) next_ps_list
           next_state = state { sas_planned_ps_list = next_ps_list }
       swapMVarF2_ result_var (next_state, real_plans)
       hPutStrLn stderr $ "Cop plan: depth " ++ show depth ++ " done"
       hPutStrLn stderr $ "Cop plan is: " ++ show (nextScore, nextMove)
       --hPutStrLn stderr $ "Cop plan all is: " ++ show nextAll
       cop_search_plan_iterate state result_var ss depth

cop_search_vote :: StepAI SearchAIState [Plan] Vote
cop_search_vote state plans result_var =
    do let ws = sas_skeleton state
           my_name = ws_my_name ws
       vote <- random_list_perm (ws_cops ws)
       let vote' = [my_name] ++
                   [name |
                    name <- vote,
                    name /= my_name]
       putMVarF2 result_var (state, vote')

cop_search_move :: StepAI SearchAIState Winner PlayerState
cop_search_move state winner result_var =
    do -- seed with identity move in case time runs out
       putMVarF2 result_var (state, sas_current_ps state)
       hPutStrLn stderr $ "Cop move: identity"
       let skel = sas_skeleton state
           update = sas_update state
           planned_ps_list = sas_planned_ps_list state
           my_name = ws_my_name skel
           cops = ws_cops skel
           Just idx = elemIndex my_name cops
           next_ps = planned_ps_list !! idx
           next_state = state { sas_current_ps = next_ps }
       swapMVarF2_ result_var (next_state, next_ps)
       hPutStrLn stderr $ "Cop move: planned"
