module Minimax where

import Evaluate
import State
import SearchState
import SearchMove
import Data.List
import Debug.Trace

type Depth = Int

minimax :: Depth -> Bool -> WorldSkeleton -> SearchState ->
           (EvalScore, [SearchMove])
minimax depth wantMax ws ss =
    case evaluateGameOver ws ss
    of
    Just gameOverEvalScore -> (gameOverEvalScore, [])
    Nothing ->
        case depth
        of
        0 -> (evaluate ws ss, [])
        _ -> minimax' depth wantMax ws ss

minimax' = minimaxRefine -- minimaxAll

minimaxAll depth wantMax ws ss =
    let allMoves = ssLegalMoves ws ss in
        minimax'' depth wantMax ws ss allMoves

minimaxRefine depth wantMax ws ss =
    let initialMoves = ssInitialMoves ws ss
        result = minimax'' depth wantMax ws ss initialMoves in
        minimaxRefine' depth wantMax ws ss result

minimaxRefine' depth wantMax ws ss result =
    --trace (show result) $
    let (bestScore, (bestMove:bestMoves)) = result
        refinedMoves = ssRefineMoves ws ss bestMove in
        case refinedMoves
        of
        [] -> result
        _ -> let result'@(bestScore', (bestMove':bestMoves')) =
                     minimax'' depth wantMax ws ss refinedMoves
                 betterThan = (if wantMax then (>) else (<)) in
                 if (bestScore' `betterThan` bestScore) then
                    minimaxRefine' depth wantMax ws ss result' else
                    result

minimax'' depth wantMax ws ss moves =
    let allStates = map (ssAdvance ws ss) moves
        allScoreNextMovePairs =
            map (minimax (depth-1) (not wantMax) ws) allStates
        allScores = map fst allScoreNextMovePairs
        allNextMoves = map snd allScoreNextMovePairs
        bestScore = (if wantMax then maximum else minimum) allScores
        Just idx = elemIndex bestScore allScores
        bestMove = moves !! idx
        bestNextMoves = allNextMoves !! idx
    in  (bestScore, (bestMove:bestNextMoves))
