module Parser where

import List
import qualified Data.Graph as G
import Text.ParserCombinators.Parsec
import Text.ParserCombinators.Parsec.Token
import Data.Graph.Inductive
import RobberTrackingImpl
import State
import PlotCourse

-- XXX should keywords be reserved?

-- winner message --

p_winner :: Parser Winner
p_winner  =  (p_keyword "winner:" >> p_name >>= return . Just)
         <|> (p_keyword "nowinner:" >> return Nothing)

-- plan messages --

p_from_plans :: Parser [Plan]
p_from_plans = p_froms p_plans >>= return . concat

p_plans :: PlayerName -> Parser [Plan]
p_plans source =
    do p_keyword "plan\\"
       plans <- many (p_plan source)
       p_keyword "plan/"
       return plans

p_plan :: PlayerName -> Parser Plan
p_plan source =
    do p_keyword "plan:"
       who <- p_name <?> "plan protagonist"
       dest <- p_name <?> "plan destination"
       job <- p_player_type <?> "plan mode of transportation"
       epoch <- p_number <?> "plan epoch"
       return (Plan { pl_source = source, pl_epoch = epoch, pl_who = who,
                      pl_move = PlayerState { ps_loc = dest, ps_type = job }})

-- inform messages --

p_from_informs :: Parser [Inform]
p_from_informs = p_froms p_informs >>= return . concat

p_informs :: PlayerName -> Parser [Inform]
p_informs source =
    do p_keyword "inf\\"
       informs <- many (p_inform source)
       p_keyword "inf/"
       return informs

p_inform :: PlayerName -> Parser Inform
p_inform source =
    do p_keyword "inf:"
       who <- p_name <?> "player name"
       loc <- p_name <?> "player location"
       job <- p_player_type <?> "player type"
       epoch <- p_number <?> "epoch"
       certainty <- p_number <?> "certainty"
       return (Inform { if_source = source, if_who = who, if_loc = loc,
                        if_ptype = job, if_epoch = epoch,
                        if_certainty = certainty })
       
-- wrapper for messages from other players --

p_froms :: (PlayerName -> Parser msg_t) -> Parser [msg_t]
p_froms p_msg =
    do p_keyword "from\\"
       msgs <- many (p_from p_msg)
       p_keyword "from/"
       return msgs

p_from :: (PlayerName -> Parser msg_t) -> Parser msg_t
p_from p_msg =
    do p_keyword "from:"
       source <- p_name <?> "message sender"
       p_msg source <?> "message body"


-- world update --

p_world_update :: Parser WorldUpdate
p_world_update =
    do p_keyword "wor\\"
       epoch <- (p_keyword "wor:" >> p_number) <?> "epoch"
       loot <- (p_keyword "rbd:" >> p_number) <?> "loot"
       banks <- p_banks <?> "banks"
       evidence <- p_evidence <?> "evidence"
       smell <- (p_keyword "smell:" >> p_number) <?> "smell"
       players <- p_players
       p_keyword "wor/"
       return (WorldUpdate { wu_epoch = epoch, wu_loot = loot,
                             wu_bank_values = banks,
                             wu_smell = if smell == 0 then Nothing
                                                      else Just smell,
                             wu_evidence = evidence, wu_players = players })

p_banks :: Parser [(LocName, Int)]
p_banks =
    do p_keyword "bv\\"
       banks <- many p_bank
       p_keyword "bv/"
       return banks

p_bank :: Parser (LocName, Int)
p_bank =
    do p_keyword "bv:"
       loc <- p_name <?> "bank location"
       val <- p_number <?> "bank value"
       return (loc, val)

p_evidence :: Parser [(LocName, Int)]
p_evidence =
    do p_keyword "ev\\"
       evidence <- many p_evidence_piece
       p_keyword "ev/"
       return evidence

p_evidence_piece :: Parser (LocName, Int)
p_evidence_piece =
    do p_keyword "ev:"
       loc <- p_name <?> "evidence location"
       world <- p_number <?> "evidence world #"
       return (loc, world)

p_players :: Parser [(PlayerName, PlayerState)]
p_players =
    do p_keyword "pl\\"
       players <- many p_player
       p_keyword "pl/"
       return players

p_player :: Parser (PlayerName, PlayerState)
p_player =
    do p_keyword "pl:"
       name <- p_name <?> "player name"
       loc <- p_name <?> "player location"
       job <- p_player_type <?> "player type"
       return (name, PlayerState { ps_loc = loc, ps_type = job })

p_player_type :: Parser PlayerType
p_player_type  =  (p_keyword "cop-foot" >> return PT_cop_foot)
              <|> (p_keyword "cop-car" >> return PT_cop_car)
              <|> (p_keyword "robber" >> return PT_robber)

-- world state --

p_world_skeleton :: Parser WorldSkeleton
p_world_skeleton =
    do p_keyword "wsk\\"
       player_name <- (p_keyword "name:" >> p_name) <?> "player name"
       robber <- (p_keyword "robber:" >> p_name) <?> "robber name"
       cops <- many ((p_keyword "cop:" >> p_name) <?> "cop name")
       locations <- p_locs <?> "locations"
       streets <- p_streets <?> "streets"
       p_keyword "wsk/"
       let hq = case [loc_name loc | loc <- locations, loc_tag loc == LT_hq] of
                [name] -> name
                [] -> error "ERROR: no headquarters"
                hqs -> error ("ERROR: mulitple headquarters" ++ show hqs)
           banks = [loc_name loc | loc <- locations, loc_tag loc == LT_bank]
           (map_player_state, ps_to_node, node_to_ps) =
               build_player_state_map locations streets hq
           robber_start = case [loc_name loc | loc <- locations,
                                    loc_tag loc == LT_rob_start] of
                [name] -> name
                [] -> error "ERROR: no robber start point"
                rs -> error ("ERROR: mulitple robber start points" ++ show rs)
           ws = WorldSkeleton { ws_my_name = player_name,
                                ws_cops = cops, ws_robber = robber,
                                ws_banks = banks, ws_hq = hq,
                                ws_robber_start = robber_start,
                                ws_ps_to_node = ps_to_node,
                                ws_node_to_ps = node_to_ps,
                                ws_map_ps = map_player_state,
                                ws_all_robber_locs = all_robber_locs,
                                ws_all_robber_loc_names = all_robber_loc_names,
                                ws_loc_id_to_name = loc_id_to_name,
                                ws_loc_name_to_id = loc_name_to_id,
                                ws_ps_to_locs_dist_1 =
                                    compute_ps_to_locs_dist 1 ws,
                                ws_ps_to_locs_dist_2 =
                                    compute_ps_to_locs_dist 2 ws,
                                ws_plot_course =
                                    compute_ws_plot_course ws}
           (all_robber_locs, all_robber_loc_names,
            loc_id_to_name, loc_name_to_id) = computeAllRobberLocs ws
       return ws

-- locations --

p_loc_type :: Parser LocType
p_loc_type  =  (p_keyword "hq" >> return LT_hq)
           <|> (p_keyword "bank" >> return LT_bank)
           <|> (p_keyword "robber-start" >> return LT_rob_start)
           <|> (p_keyword "ordinary" >> return LT_ord)

p_loc :: Parser Loc
p_loc = do p_keyword "nod:"
           name <- p_name <?> "location name"
           tag <- p_loc_type <?> "location tag"
           coord_x <- p_number <?> "x-coord"
           coord_y <- p_number <?> "y-coord"
           return (Loc { loc_name = name, loc_tag = tag,
                         loc_coords = (coord_x, coord_y) })


p_locs :: Parser [Loc]
p_locs = do p_keyword "nod\\"
            locs <- many p_loc <?> "locations"
            p_keyword "nod/"
            return locs

-- streets --

p_car_only :: Parser Bool
p_car_only  =  (p_keyword "car" >> return True)
              <|> (p_keyword "foot" >> return False)

p_street :: Parser Street
p_street = do p_keyword "edg:"
              start <- p_name <?> "street start location"
              end <- p_name <?> "street end location"
              car_only <- p_car_only <?> "street type"
              return (Street { st_start = start, st_end = end,
                               st_car_only = car_only })


p_streets :: Parser [Street]
p_streets = do p_keyword "edg\\"
               streets <- many p_street
               p_keyword "edg/"
               return streets

-- general --

p_name :: Parser String
p_name = identifier robber_tokenizer

p_keyword :: String -> Parser ()
p_keyword = reserved robber_tokenizer

p_number :: Parser Int
p_number =
    do negative <- option False (char '-' >> return True)
       natural <- fmap fromInteger (natural robber_tokenizer) <?> "number"
       return (if negative then -natural else natural)

p_or_game_over :: Parser some_t -> Parser (Maybe some_t)
p_or_game_over parser  =  (p_keyword "game-over" >> return Nothing)
                      <|> (parser >>= return . Just)
--- tokenizer ---

robber_language_def :: LanguageDef parse_state
robber_language_def =
     LanguageDef { commentStart = "/*",
                   commentEnd = "*/",
                   commentLine = ";",
                   nestedComments = True,
                   identStart = letter <|> digit <|> oneOf "-_#()",
                   identLetter = letter <|> digit <|> oneOf "-_#()",
                   opStart = fail "no operators",
                   opLetter = fail "no operators",
                   reservedNames = [-- player type
                                    "cop-foot", "cop-car", "robber",
                                    -- edge type
                                    "car", "foot",
                                    -- node tag
                                    "hq", "bank", "robber-start", "ordinary",
                                    -- keywords
                                    "wsk\\", "wsk/", "nod\\", "nod/",
                                    "edg\\", "edg/",
                                    "name:", "robber:", "cop:",
                                    "nod:", "edg:"],
                   reservedOpNames = ["-"],
                   caseSensitive = False }

robber_tokenizer :: TokenParser parse_state
robber_tokenizer = makeTokenParser robber_language_def

