module Protocol where

import System
import IO
import Monad
import Text.ParserCombinators.Parsec
import Control.Concurrent
import Control.Concurrent.MVar
import System.Posix.Signals
import CPUTime
import Numeric

import State
import Parser
import AI

timeout_in_s :: Int
timeout_in_s = 4

enable_watchdog :: Bool
enable_watchdog = True

run_robber :: RobberAI state_t -> IO ()
run_robber ai =
    do state_var <- newEmptyMVar
       state_result_var <- newEmptyMVar
       send_reg (rai_name ai) PT_robber
       skel <- read_skeleton
       state <- run_step ((rai_init ai) skel) state_var (rai_name ai ++ " init")
       run_robber_loop ai state state_result_var

run_robber_loop :: RobberAI state_t -> state_t -> MVar (state_t, PlayerState)
                -> IO ()
run_robber_loop ai state move_var =
    do let name = rai_name ai
       update <- read_update
       (next_state, next_move) <-
           run_step ((rai_move ai) state update) move_var
                    (rai_name ai ++ " move")
       send_move next_move
       run_robber_loop ai next_state move_var

run_cop :: CopAI state_t -> IO ()
run_cop ai =
    do state_var <- newEmptyMVar
       inform_var <- newEmptyMVar
       plan_var <- newEmptyMVar
       vote_var <- newEmptyMVar
       move_var <- newEmptyMVar
       send_reg (cai_name ai) (cai_init_type ai)
       skel <- read_skeleton
       state <- run_step ((cai_init ai) skel) state_var (cai_name ai ++ " init")
       run_cop_loop ai state inform_var plan_var vote_var move_var

run_cop_loop :: CopAI state_t -> state_t
             -> MVar (state_t, [Inform]) -> MVar (state_t, [Plan])
             -> MVar (state_t, Vote) -> MVar (state_t, PlayerState)
             -> IO ()
run_cop_loop ai state0 inform_var plan_var vote_var move_var =
    do let sn stage = cai_name ai ++ (' ' : stage)
       update <- read_update
       (state1, my_informs) <-
           run_step ((cai_inform ai) state0 update) inform_var (sn "inform")
       send_informs my_informs
       informs <- read_informs
       (state2, my_plans) <-
           run_step ((cai_plan ai) state1 informs) plan_var (sn "plan")
       send_plans my_plans
       plans <- read_plans
       (state3, my_vote) <-
           run_step ((cai_vote ai) state2 plans) vote_var (sn "vote")
       send_vote my_vote
       winner <- read_winner
       (state4, my_move) <-
           run_step ((cai_move ai) state3 winner) move_var (sn "move")
       send_move my_move
       run_cop_loop ai state4 inform_var plan_var vote_var move_var

run_step :: (MVar some_t -> IO ()) -> MVar some_t -> String -> IO some_t
run_step step var stage_name =
    do _ <- if enable_watchdog then scheduleAlarm timeout_in_s else return 0
       timeout_var <- newEmptyMVar
       let alarm = CatchOnce (putMVar timeout_var True)
       old_alarm <- installHandler realTimeAlarm alarm Nothing
       time_start <- getCPUTime
       step_thread_id <- forkIO (step var >> putMVar timeout_var False)
       timed_out <- takeMVar timeout_var
       old_alarm <- installHandler realTimeAlarm alarm Nothing
       _ <- installHandler realTimeAlarm old_alarm Nothing
       _ <- if enable_watchdog then scheduleAlarm 0 else return 0
       when timed_out (killThread step_thread_id)
       time_end <- getCPUTime
       let time_diff =
             " (" ++ showFFloat (Just 4)
                     ((fromInteger (time_end - time_start))/1.0e12) " sec)"
       hPutStrLn stderr ("STATS: " ++ stage_name ++
                         if timed_out then " (timed out)" else time_diff)
       take_mvar_or_die var stage_name

take_mvar_or_die :: MVar some_t -> String -> IO (some_t)
take_mvar_or_die var stage_name =
    tryTakeMVar var >>= \contents ->
    case contents of
    Just result -> return result
    Nothing -> do hPutStrLn stderr
                      ("ERROR: stage " ++ show stage_name ++
                       " did not compute a result")
                  exitWith (ExitFailure 3)

send_reg :: String -> PlayerType -> IO ()
send_reg name player_type =
    do hSetBuffering stdin LineBuffering
       hSetBuffering stdout LineBuffering
       putStr "reg: "
       putStr name
       putChar ' '
       putStrLn (message player_type)

send_move :: PlayerState -> IO ()
send_move player_state =
    do putStr "mov: "
       putStrLn (message player_state)

send_informs :: [Inform] -> IO ()
send_informs infs =
    do putStrLn "inf\\"
       mapM_ (putStrLn . ("inf: " ++) . message) infs
       putStrLn "inf/"
      
send_plans :: [Plan] -> IO ()
send_plans plans =
    do putStrLn "plan\\"
       mapM_ (putStrLn . ("plan: " ++) . message) plans
       putStrLn "plan/"

send_vote :: [PlayerName] -> IO ()
send_vote names =
    do putStrLn "vote\\"
       mapM_ (putStrLn . ("vote: " ++)) names
       putStrLn "vote/"
      
read_skeleton :: IO WorldSkeleton
read_skeleton = parse_from_stdin p_world_skeleton (== "wsk/")

read_update :: IO WorldUpdate
read_update = parse_from_stdin p_world_update (== "wor/")

read_informs :: IO [Inform]
read_informs = parse_from_stdin p_from_informs (== "from/")

read_plans :: IO [Plan]
read_plans = parse_from_stdin p_from_plans (== "from/")

read_winner :: IO Winner
read_winner = parse_from_stdin p_winner (const True)

parse_from_stdin :: Parser some_t -> (String -> Bool) -> IO some_t
parse_from_stdin parser is_end =
    do lines <- read_lines_until is_end
       case runParser (p_or_game_over parser) () "<stdin>" (unlines lines) of
         Left err -> print err >> exitWith (ExitFailure 2)
         Right Nothing -> exitWith ExitSuccess
         Right (Just result) -> return result

read_lines_until :: (String -> Bool) -> IO [String]
read_lines_until is_end =
    do at_eof <- isEOF
       when at_eof (exitWith ExitSuccess)
       line <- getLine
       if is_end line
         then return [line]
         else read_lines_until is_end >>= return . (line :)

