
module RobberHeuristic where

import qualified Data.Set as S
import Data.List
import Control.Concurrent.MVar
import IO
import Debug.Trace

import ConcurrentUtil
import State
import AI
import LegalMoves
import RandomUtils
import PlotCourse
import GraphUtil
import RobberTrackingImpl

-- robber-heuristic goes for the richest bank, avoiding cops

-- ai config --

-- cost for distance-from-cop: 0, 1, ... in cash ("will risk it for $x") units
cost_per_dist :: [Int]
cost_per_dist = [100000, 10000, 1000, 100, 10] ++ repeat 1

max_radius :: Int
max_radius = length (takeWhile (> 1) cost_per_dist)

-- higher score = better chance of accepting plan
score :: (Int, Distance, LocName, [PlayerState]) -> Int
score inp@(cash, dist, locnam, path) =
    let final_score = cash - dist
--    in  trace ("SCORE: " ++ show final_score ++ " for " ++ locnam ++ " with $" ++ show cash ++ " at " ++ show dist) final_score
    in  final_score

-- ai implementation --

robber_heuristic = RobberAI { rai_name = "robber-heuristic",
                              rai_init = robber_heuristic_init,
                              rai_move = robber_heuristic_move }

data HeuristicState = HeuristicState {
    ss_ws :: WorldSkeleton,
    ss_current_ps :: PlayerState
}

robber_heuristic_init :: InitAI HeuristicState
robber_heuristic_init skel state_var =
    putMVar state_var
        (HeuristicState { ss_ws = skel,
                         ss_current_ps = PlayerState {
                                             ps_type = PT_robber,
                                             ps_loc = ws_robber_start skel }})

robber_heuristic_move :: StepAI HeuristicState WorldUpdate PlayerState
robber_heuristic_move state update result_var =
    do -- seed with identity move in case time runs out
       let here = ss_current_ps state
           ws = ss_ws state
       -- hPutStrLn stderr $ "RH: identity"
       putMVarF2 result_var (state, here)
       let cops = [ps | (_, ps) <- wu_players update, ps_type ps /= PT_robber]
           radii_maps = [(r, map_avoiding_cops ws cops r) | r <- [0..]]
       let plans_sets =
               [(care_radius,
                 [(val, cost, loc, path)
                  | (loc,val) <- wu_bank_values update,
                    let (path,cost) = plot_course_with_map ws pmap here loc])
                 | (care_radius,pmap) <- take max_radius radii_maps]
       mapM_ (robber_heuristic_move_path_set state result_var) plans_sets

robber_heuristic_move_path_set state result_var (radius,plans) =
    do -- hPutStrLn stderr $ "RH: radius " ++ show radius
       let ((scr, path, bank) : _) = sort [(-(score plan), path, loc)
                                         | plan@(val, len, loc, path) <- plans]
           move = case path of
                  (here : next : _) -> next
                  [here] -> here
                  [] -> error "robber_heuristic_move_path_set: phantom bank"
           next_state = state { ss_current_ps = move }
       bank `seq` scr `seq` return ()
       -- hPutStrLn stderr $ "RH: score for " ++ bank ++ " = " ++ show scr
       -- hPutStrLn stderr $ "RH: move towards " ++ bank ++ " = " ++ show move
       -- hPutStrLn stderr $ "RH: path: " ++ show path
       swapMVarF2_ result_var (next_state, move)

-- robber-nodes at various distances from cops
nodesets_near_cops ws = remove_dupls . combine_node_sets . map (nodesets_near_cop ws)

-- robber-nodes at various distances from each cop
nodesets_near_cop ws cop_ps =
    let robbify ps = ps { ps_type = PT_robber }
        ps_to_r_node = ws_ps_to_node ws . robbify
        nss = remove_dupls $ map S.fromList $
            [ps_to_r_node cop_ps] :
            map (map ps_to_r_node . computeStatesAtDistance ws cop_ps) [1..]
    in  nss

map_avoiding_cops ws cops care_radius =
    let nss = take (care_radius + 1) (nodesets_near_cops ws cops)
        cost_dsts = concat [[(n,cst) | n <- S.toList ns]
                            | (ns,cst) <- zip nss cost_per_dist]
        mk_msg (n, cst) = "RH: costs: $" ++ show cst ++
                           " for " ++ show ((ws_node_to_ps ws) n,n) ++ "\n"
        new_map = repl_edges_to (ws_map_ps ws) cost_dsts
        msg1 = concatMap mk_msg cost_dsts
        msg2 = "RH: map:\n" ++ show new_map
--    in  trace msg1 $ trace msg2 $ new_map
    in  new_map

-- remove elements in earlier sets from later sets
remove_dupls :: (Ord elt_t) => [S.Set elt_t] -> [S.Set elt_t]
remove_dupls sets = remove_dupls' sets S.emptySet 

remove_dupls' :: (Ord elt_t) => [S.Set elt_t] -> S.Set elt_t -> [S.Set elt_t]
remove_dupls' [] _ = []
remove_dupls' (s:ss) seen = (s `S.minusSet` seen) : remove_dupls' ss (seen `S.union` s)

combine_node_sets :: (Ord elt_t) => [[S.Set elt_t]] -> [S.Set elt_t]
combine_node_sets setss =
    S.unions (map head setss) : combine_node_sets (map tail setss)

