
module RobberJanosik where

import qualified Data.Set as S
import Data.List
import Control.Concurrent.MVar
import IO
import Debug.Trace
import Monad
import Data.Graph.Inductive

import ConcurrentUtil
import State
import AI
import LegalMoves
import RandomUtils
import PlotCourse
import GraphUtil
import RobberTrackingImpl

-- robber-janosik goes for the richest bank, avoiding cops, and then hides
-- for a while

-- ai config --

-- cost for distance-from-cop: 0, 1, ... in cash ("will risk it for $x") units
cost_per_dist :: [Int]
cost_per_dist = [100000, 10000, 1000, 100, 10] ++ repeat 1

max_radius :: Int
max_radius = length (takeWhile (> 1) cost_per_dist)

hideout_safety = 5
num_hideouts = 10
max_heat = 6

-- higher score = better chance of accepting plan
score :: (Int, Distance, LocName, [PlayerState]) -> Int
score inp@(cash, dist, locnam, path) =
    let final_score = cash - dist
--    in  trace ("SCORE: " ++ show final_score ++ " for " ++ locnam ++ " with $" ++ show cash ++ " at " ++ show dist) final_score
    in  final_score

-- ai implementation --

robber_janosik = RobberAI { rai_name = "janosik",
                            rai_init = robber_janosik_init,
                            rai_move = robber_janosik_move }

data JanosikState = JanosikState {
    js_ws :: WorldSkeleton,
    js_heat :: Int,
    js_hideouts :: [LocId],
    js_banks :: S.Set PlayerState,
    js_current_ps :: PlayerState
}

robber_janosik_init :: InitAI JanosikState
robber_janosik_init skel state_var =
    putMVar state_var
        (JanosikState { js_ws = skel,
                        js_heat = max_heat,
                        js_hideouts = [],
                        js_banks = S.fromList [PlayerState { ps_loc = bank,
                                                      ps_type = PT_robber }
                                                  | bank <- ws_banks skel],
                        js_current_ps = PlayerState {
                                             ps_type = PT_robber,
                                             ps_loc = ws_robber_start skel }})

robber_janosik_move :: StepAI JanosikState WorldUpdate PlayerState
robber_janosik_move state update result_var =
    do -- seed with identity move in case time runs out
       let here = js_current_ps state
           ws = js_ws state
       -- hPutStrLn stderr $ "RH: identity"
       putMVarF2 result_var (state, here)
       let heat = js_heat state
           state_less_heat = state { js_heat = heat - 1 }
       -- hPutStrLn stderr ("HEAT: " ++ show heat)
       if (heat > 0)
          then robber_janosik_move_hide state_less_heat update result_var
          else robber_janosik_move_rob state update result_var

-- if the heat is on, hide
robber_janosik_move_hide :: StepAI JanosikState WorldUpdate PlayerState
robber_janosik_move_hide state update result_var =
    do let here = js_current_ps state
           ws = js_ws state
           cops = [ps | (_, ps) <- wu_players update, ps_type ps /= PT_robber]
           radii_maps = [(r, map_avoiding_cops ws cops r) | r <- [0..]]
       hideouts <- update_hideouts state cops
       let best_hide_path (radius, hps) =
               let ((_, path):_) = sort hps in (radius, path)
           plans_sets = [(care_radius, [(cost, path) | n <- hideouts, let (path, cost) = plot_course_with_map ws pmap here (ps_loc ((ws_node_to_ps ws) n))])
                      | (care_radius, pmap) <- take max_radius radii_maps]
       mapM_ (robber_janosik_move_path_set state result_var . best_hide_path)
             plans_sets

update_hideouts :: JanosikState -> [PlayerState] -> IO [LocId]
update_hideouts state cops =
    do let ws = js_ws state
           hot_set = S.unions $ take hideout_safety
                     (S.map (ws_ps_to_node ws) (js_banks state) :
                      nodesets_near_cops ws cops)
           safe_hides = [h | h <- js_hideouts state,
                             not (h `S.member` hot_set)]
           n_more_hides = num_hideouts - length safe_hides
           cold_set = ws_all_robber_locs ws `S.difference` hot_set
           n_cold_set = S.size cold_set
           pick_hide (places, new_hides) _ =
               do hide <- random_list_elt places
                  return (delete hide places, hide : new_hides)
       (_, new_hides) <- foldM pick_hide (S.toList cold_set, safe_hides)
                               [0 .. (num_hideouts - length safe_hides) - 1]
       return new_hides


-- when the heat is not on, go rob something
robber_janosik_move_rob :: StepAI JanosikState WorldUpdate PlayerState
robber_janosik_move_rob state update result_var =
    do let here = js_current_ps state
           ws = js_ws state
           cops = [ps | (_, ps) <- wu_players update, ps_type ps /= PT_robber]
           radii_maps = [(r, map_avoiding_cops ws cops r) | r <- [0..]]
           best_bank_path (radius, bps) =
               let ((_, path, _):_) = sort [(-(score plan), path, loc)
                                        | plan@(val, len, loc, path) <- bps]
               in  (radius, path)
           plans_sets =
               [(care_radius,
                 [(val, cost, loc, path)
                  | (loc,val) <- wu_bank_values update,
                    let (path,cost) = plot_course_with_map ws pmap here loc])
                 | (care_radius,pmap) <- take max_radius radii_maps]
       mapM_ (robber_janosik_move_path_set state result_var . best_bank_path)
             plans_sets

robber_janosik_move_path_set state result_var (radius,path) =
    do -- hPutStrLn stderr $ "RH: radius " ++ show radius
       let move = case path of
                  (here : next : _) -> next
                  [here] -> here
                  [] -> error "robber_janosik_move_path_set: phantom bank"
           next_state | move `S.member` js_banks state =
                            state { js_current_ps = move, js_heat = max_heat }
                      | otherwise =
                            state { js_current_ps = move }
       swapMVarF2_ result_var (next_state, move)

-- robber-nodes at various distances from cops
nodesets_near_cops ws = remove_dupls . combine_node_sets . map (nodesets_near_cop ws)

-- robber-nodes at various distances from each cop
nodesets_near_cop ws cop_ps =
    let robbify ps = ps { ps_type = PT_robber }
        ps_to_r_node = ws_ps_to_node ws . robbify
        nss = remove_dupls $ map S.fromList $
            [ps_to_r_node cop_ps] :
            map (map ps_to_r_node . computeStatesAtDistance ws cop_ps) [1..]
    in  nss

map_avoiding_cops ws cops care_radius =
    let nss = take (care_radius + 1) (nodesets_near_cops ws cops)
        cost_dsts = concat [[(n,cst) | n <- S.toList ns]
                            | (ns,cst) <- zip nss cost_per_dist]
        mk_msg (n, cst) = "RH: costs: $" ++ show cst ++
                           " for " ++ show ((ws_node_to_ps ws) n,n) ++ "\n"
        new_map = repl_edges_to (ws_map_ps ws) cost_dsts
        msg1 = concatMap mk_msg cost_dsts
        msg2 = "RH: map:\n" ++ show new_map
--    in  trace msg1 $ trace msg2 $ new_map
    in  new_map

-- remove elements in earlier sets from later sets
remove_dupls :: (Ord elt_t) => [S.Set elt_t] -> [S.Set elt_t]
remove_dupls sets = remove_dupls' sets S.emptySet 

remove_dupls' :: (Ord elt_t) => [S.Set elt_t] -> S.Set elt_t -> [S.Set elt_t]
remove_dupls' [] _ = []
remove_dupls' (s:ss) seen = (s `S.difference` seen) : remove_dupls' ss (seen `S.union` s)

combine_node_sets :: (Ord elt_t) => [[S.Set elt_t]] -> [S.Set elt_t]
combine_node_sets setss =
    S.unions (map head setss) : combine_node_sets (map tail setss)

