module RobberSearch where

import IO
import Control.Concurrent.MVar
import Data.Maybe

import ConcurrentUtil
import State
import AI
import LegalMoves
import RandomUtils
import SearchState
import SearchMove
import Minimax
import Evaluate
import RobberTracking
import RobberTrackingImpl

-- robber-search makes moves using a game tree search

robber_search = RobberAI { rai_name = "robber-search",
                           rai_init = robber_search_init,
                           rai_move = robber_search_move }

data SearchAIState = SearchAIState {
    sas_skeleton :: WorldSkeleton,
    sas_current_ps :: PlayerState,
    sas_impossible_list :: [LocSet]
}

robber_search_init :: InitAI SearchAIState
robber_search_init skel state_var =
    putMVar state_var
        (SearchAIState { sas_skeleton = skel,
                         sas_current_ps = PlayerState {
                                             ps_type = PT_robber,
                                             ps_loc = ws_robber_start skel },
                         sas_impossible_list =
                             initImpossibleList skel })

robber_search_move :: StepAI SearchAIState WorldUpdate PlayerState
robber_search_move state update result_var =
    do -- seed with identity move in case time runs out
       putMVarF2 result_var (state, sas_current_ps state)
       hPutStrLn stderr $ "Robber move: identity"
       let skel = sas_skeleton state
           prevImpossibleList = sas_impossible_list state
           ss = SearchState { ss_epoch = wu_epoch update,
                              ss_cop_states =
                                  map (fromJust .
                                       ((flip lookup) (wu_players update)))
                                      (ws_cops skel),
                              ss_robber_state =
                                  fromJust $
                                  lookup (ws_robber skel)
                                         (wu_players update),
                              ss_fake_robber_states = Nothing,
                              ss_impossible_list = prevImpossibleList }
       robber_search_move_iterate state update result_var ss 0

robber_search_move_iterate state update result_var ss prevDepth =
    do let skel = sas_skeleton state
           depth = nextDepth skel ss prevDepth
       hPutStrLn stderr $ "Robber move: depth " ++ show depth ++ " start"
       let (nextScore, (nextMove@(RobberMove next_ps):_)) =
               minimax depth False skel ss
           ss' = ssAdvance skel ss nextMove
           next_state = state { sas_current_ps = next_ps,
                                sas_impossible_list = ss_impossible_list ss' }
       swapMVarF2_ result_var (next_state, next_ps)
       hPutStrLn stderr $ "Robber move: depth " ++ show depth ++ " done"
       hPutStrLn stderr $ "Robber move is: " ++ show (nextScore, nextMove)
       robber_search_move_iterate state update result_var ss depth
