module RobberTrackingImpl where

import State
import LegalMoves
import Data.List
import Data.Graph.Inductive
import SearchState
import RobberTracking
import Data.Maybe
import qualified Data.Map as M


allRobberLocs :: WorldSkeleton -> LocSet
allRobberLocs ws = ws_all_robber_locs ws

computeAllRobberLocNames ws =
    let robber_start_node = (ws_ps_to_node ws)
                            (PlayerState { ps_loc = (ws_robber_start ws),
                                           ps_type = PT_robber })
    in  map (ps_loc . (ws_node_to_ps ws))
            (reachable robber_start_node (ws_map_ps ws))

computeAllRobberLocs ws =
    let names = computeAllRobberLocNames ws
        nameIdPairs = zip names [0..]
        idNamePairs = zip [0..] names
        ids = map fst idNamePairs
        nameIdMap = M.fromList nameIdPairs
        idNameMap = M.fromList idNamePairs
        nameToId = \ name -> fromJust (M.lookup name nameIdMap)
        idToName = \ id -> fromJust (M.lookup id idNameMap) in
        (lsFromList ids, names, idToName, nameToId)

initImpossibleList ws = [(allRobberLocs ws)
                         `lsDelete`
                         ((ws_loc_name_to_id ws) (ws_robber_start ws))]

copCurImpossibleLocs ws wu copName =
    lsUnions
    [copCurImpossibleLocsCops ws wu,
     copCurImpossibleLocsRobbing ws wu,
     copCurImpossibleLocsSense ws wu copName
    --,copCurImpossibleLocsEvidence ws wu
    ]

copCurImpossibleLocsCops ws wu =
    let cops = ws_cops ws
        copStates =
            map
            (fromJust . ((flip lookup) (wu_players wu)))
            cops
        copLocs =
            lsFromList (map ((ws_loc_name_to_id ws) . stateToLoc) copStates) in
        copLocs

copCurImpossibleLocsRobbing ws wu =
    case lookup (ws_robber ws) (wu_players wu)
    of
    Just robberState ->
        (allRobberLocs ws)
        `lsDelete`
        ((ws_loc_name_to_id ws) (stateToLoc robberState))
    Nothing -> lsEmpty

copCurImpossibleLocsSense ws wu copName =
    let robberName = ws_robber ws
        Just copPlayerState = lookup copName (wu_players wu) in
        case wu_smell wu
        of
        Just dist ->
            (allRobberLocs ws)
            `lsDifference`
            (case dist
             of
             1 -> ((ws_ps_to_locs_dist_1 ws) copPlayerState)
             2 -> ((ws_ps_to_locs_dist_2 ws) copPlayerState))
        Nothing ->
            case ps_type copPlayerState
            of
            PT_cop_car ->
                ((ws_ps_to_locs_dist_1 ws) copPlayerState)
            PT_cop_foot ->
                ((ws_ps_to_locs_dist_1 ws) copPlayerState) `lsUnion`
                ((ws_ps_to_locs_dist_2 ws) copPlayerState)

copCurImpossibleLocsEvidence ws wu =
    let evList = wu_evidence wu in
        lsUnions $
        map (copCurImpossibleLocsEvidence' ws wu) evList

copCurImpossibleLocsEvidence' ws wu (evLocName, evEpoch) =
    let myLocName =
            stateToLoc $ fromJust $ lookup (ws_my_name ws) (wu_players wu)
        curEpoch = wu_epoch wu
        steps = (curEpoch - evEpoch - 1) `div` 2 in
        if evEpoch >= 8 &&
           evEpoch `mod` 8 == 0 &&
           --steps <= 11 &&
           evLocName == myLocName
        then let evLoc = (ws_loc_name_to_id ws) evLocName
                 seedPossibleLocs = lsSingleton evLoc
                 iteratedPossibleLocs =
                     (iterate (advancePossibleLocs ws) seedPossibleLocs)
                     !! steps in
                 (ws_all_robber_locs ws) `lsDifference` iteratedPossibleLocs
        else lsEmpty
           
ssCurImpossibleLocs ws ss =
    lsUnions
    [ssCurImpossibleLocsCops ws ss,
     ssCurImpossibleLocsRobbing ws ss,
     ssCurImpossibleLocsSense ws ss]

ssCurImpossibleLocsCops ws ss =
    let robberLoc = (ws_loc_name_to_id ws) (stateToLoc (ss_robber_state ss))
        copStates = ss_cop_states ss
        copLocs =
            lsFromList (map ((ws_loc_name_to_id ws) . stateToLoc) copStates) in
        copLocs `lsDelete` robberLoc

ssCurImpossibleLocsRobbing ws ss =
    let robberLoc = stateToLoc (ss_robber_state ss)
        bankLocs = (ws_banks ws) in
        if elem robberLoc bankLocs
        then (allRobberLocs ws)
             `lsDelete`
             ((ws_loc_name_to_id ws) robberLoc)
        else lsEmpty

ssCurImpossibleLocsSense ws ss =
    let copStates = ss_cop_states ss
        robberLoc = (ws_loc_name_to_id ws) (stateToLoc (ss_robber_state ss)) in
        lsUnions
        (map (ssCurImpossibleLocsSense' ws robberLoc) copStates)

ssCurImpossibleLocsSense' ws robberLoc copState =
    let senseLocs1 = (ws_ps_to_locs_dist_1 ws) copState
        senseLocs2 = (ws_ps_to_locs_dist_2 ws) copState
    in case ps_type copState
       of
       PT_cop_car ->
           if robberLoc `lsMember` senseLocs1
           then (allRobberLocs ws) `lsDifference` senseLocs1
           else senseLocs1
       PT_cop_foot ->
           if robberLoc `lsMember` senseLocs1
           then (allRobberLocs ws) `lsDifference` senseLocs1
           else if robberLoc `lsMember` senseLocs2
                then (allRobberLocs ws) `lsDifference` senseLocs2
                else senseLocs1 `lsUnion` senseLocs2

nextImpossibleList :: WorldSkeleton -> ImpossibleList -> LocSet ->
                      ImpossibleList
nextImpossibleList ws prevImpossibleList curImpossibleLocs =
    let allLocs = allRobberLocs ws
        prevImpossibleLocs = head prevImpossibleList
        prevPossibleLocs = allLocs `lsDifference` prevImpossibleLocs
        advancedPossibleLocs = advancePossibleLocs ws prevPossibleLocs
        nextPossibleLocs =
            advancedPossibleLocs `lsDifference` curImpossibleLocs
        nextImpossibleLocs = allLocs `lsDifference` nextPossibleLocs in
        (nextImpossibleLocs:prevImpossibleList)

advancePossibleLocs :: WorldSkeleton -> LocSet -> LocSet
advancePossibleLocs ws prevPossibleLocs =
    let prevPossibleStates =
            map
            ((locToState PT_robber) . (ws_loc_id_to_name ws) )
            (lsToList prevPossibleLocs)
        advancedPossibleLocs =
            lsUnions $
            map (lsFromList .
                 (map ((ws_loc_name_to_id ws) . stateToLoc)) .
                 legalMoves ws)
                prevPossibleStates in
        advancedPossibleLocs

computeStatesUpToDistance ws ps 0 = [ps]
computeStatesUpToDistance ws ps dist =
    let prevStates = computeStatesAtDistance ws ps (dist-1) in
        computeStatesUpToDistance' ws ps prevStates

computeStatesUpToDistance' ws ps prevStates =
    nub $ concatMap
            (legalMovesSameState ws)
            prevStates

computeStatesAtDistance ws ps 0 = [ps]
computeStatesAtDistance ws ps dist =
    let upToPrevDistance = computeStatesUpToDistance ws ps (dist-1)
        upToDistance = computeStatesUpToDistance' ws ps upToPrevDistance in
        upToDistance \\ upToPrevDistance

computeLocsAtDistance ws ps dist =
    lsFromList $
    map
    ((ws_loc_name_to_id ws) . stateToLoc)
    (computeStatesAtDistance ws ps dist)

compute_ps_to_locs_dist dist ws =
    let allPlayerStates = map snd (labNodes (ws_map_ps ws))
        associations = [(playerState,
                         computeLocsAtDistance ws playerState dist) |
                        playerState <- allPlayerStates]
        theMap = M.fromList associations in
        \ playerState -> fromJust (M.lookup playerState theMap)
