module State where

import List
import qualified Data.Graph as G
import qualified Data.Map as M
import Data.Graph.Inductive
import Data.Graph.Inductive.Monad.IOArray
import Text.Show.Functions
import qualified Data.Set as Set

class Message some_t where
    message :: some_t -> String

data WorldSkeleton = WorldSkeleton {
    ws_my_name :: PlayerName,
    ws_cops :: [PlayerName],
    ws_robber :: PlayerName,
    ws_hq :: LocName,
    ws_robber_start :: LocName,
    ws_banks :: [LocName],
    ws_ps_to_node :: PlayerState -> Node,
    ws_node_to_ps :: Node -> PlayerState,
    ws_map_ps :: PlayerStateMap,
    ws_all_robber_locs :: LocSet,
    ws_all_robber_loc_names :: [LocName],
    ws_loc_id_to_name :: LocId -> LocName,
    ws_loc_name_to_id :: LocName -> LocId,
    ws_ps_to_locs_dist_1 :: PlayerStateLocSetMap,
    ws_ps_to_locs_dist_2 :: PlayerStateLocSetMap,
    ws_plot_course :: PlayerState -> LocName -> ([PlayerState], Distance)
} deriving (Show)


data WorldUpdate = WorldUpdate {
    wu_epoch :: Epoch,
    wu_loot :: Int,
    wu_bank_values :: [(LocName, Int)],
    wu_evidence :: [(LocName, Epoch)],
    wu_smell :: Maybe Distance,
    wu_players :: [(PlayerName, PlayerState)]
} deriving (Show)

data Inform = Inform {
    if_source :: !PlayerName, -- information source
    if_who :: !PlayerName,
    if_loc :: !LocName,
    if_ptype :: !PlayerType,
    if_epoch :: !Epoch,
    if_certainty :: !Certainty
} deriving (Show)

instance Message Inform where
    message inf = if_who inf ++ " " ++ if_loc inf ++ " " ++
                  message (if_ptype inf) ++ " " ++ message (if_epoch inf) ++
                  " " ++ message (if_certainty inf)

data Plan = Plan {
    pl_source :: !PlayerName,
    pl_epoch :: !Epoch,
    pl_who :: !PlayerName,
    pl_move :: !PlayerState
} deriving (Show)

instance Message Plan where
    message plan = pl_who plan ++ " " ++ message (pl_move plan) ++
                   " " ++ message (pl_epoch plan)

data LocType = LT_hq
             | LT_bank
             | LT_rob_start
             | LT_ord
    deriving (Eq, Show)

data Loc = Loc { loc_name :: LocName,
                 loc_tag :: LocType,
                 loc_coords :: (Int, Int) }
    deriving (Show)

data PlayerType = PT_cop_foot | PT_cop_car | PT_robber
    deriving (Eq, Ord)

instance Show PlayerType where
    show pt = message pt

instance Message PlayerType where
    message PT_cop_foot = "cop-foot"
    message PT_cop_car = "cop-car"
    message PT_robber = "robber"

data PlayerState = PlayerState {
    ps_loc :: !LocName,
    ps_type :: !PlayerType
} deriving (Eq, Ord)

instance Show PlayerState where
    show (PlayerState { ps_loc = loc, ps_type = pt }) =
        show pt ++ "@" ++ show loc

type PlayerStateMap = Gr PlayerState Int -- # is travel cost

stateToLoc = ps_loc
locToState playerType loc =
    PlayerState { ps_loc = loc, ps_type = playerType }

instance Message PlayerState where
    message pstate = ps_loc pstate ++ " " ++ message (ps_type pstate)

type Vote = [PlayerName]
type Winner = Maybe PlayerName

type PlayerName = String
type LocName = String
type Epoch = Int
type Distance = Int
type Certainty = Int

instance Message Int where
    message n = show n

robberToMove epoch = epoch `mod` 2 == 0
copToMove epoch = epoch `mod` 2 == 1

data StreetType = ST_car | ST_foot
    deriving (Eq, Show)

data Street = Street { st_start :: LocName,
                       st_end :: LocName,
                       st_car_only :: Bool }
    deriving (Show)

build_player_state_map :: [Loc] -> [Street] -> LocName
        -> (PlayerStateMap, PlayerState -> Node, Node -> PlayerState)
build_player_state_map locations streets hq =
    let all_states = concat
            [[PlayerState { ps_loc = l, ps_type = PT_robber }, 
              PlayerState { ps_loc = l, ps_type = PT_cop_foot }, 
              PlayerState { ps_loc = l, ps_type = PT_cop_car }]
             | loc <- locations, let l = loc_name loc]
        uniq_nodes :: [(Node, PlayerState)]
        uniq_nodes = zip [0..] all_states
        node_to_ps_map :: M.Map Node PlayerState
        node_to_ps_map = M.fromList uniq_nodes
        node_to_ps n = case M.lookup n node_to_ps_map of
                        Just ps -> ps
                        Nothing -> error ("ERROR: phantom node: " ++ show n)
        ps_to_node_map :: M.Map PlayerState Node
        ps_to_node_map = M.fromList (zip all_states [0..])
        ps_to_node ps = case M.lookup ps ps_to_node_map of
                       Just n -> n
                       Nothing -> error ("ERROR: phantom state: " ++ show ps)
        mk_edge pt start end = (PlayerState { ps_type = pt, ps_loc = start },
                                PlayerState { ps_type = pt, ps_loc = end })
        street_to_edge st | st_car_only st =
           -- cops can drive on any car-only street in the right direction
           [mk_edge PT_cop_car (st_start st) (st_end st)]
        street_to_edge st | otherwise =
           -- cops can drive on any normal street in the right direction
           [mk_edge PT_cop_car (st_start st) (st_end st),
           -- cops can walk on any normal street in any direction
            mk_edge PT_cop_foot (st_start st) (st_end st),
            mk_edge PT_cop_foot (st_end st) (st_start st),
           -- robbers can walk on any normal street in any direction
            mk_edge PT_robber (st_start st) (st_end st),
            mk_edge PT_robber (st_end st) (st_start st)]
        mk_hq_edges st | st_start st == hq && st_car_only st =
           -- cops can take a car from hq to end of any car-only street
           [(PlayerState { ps_type = PT_cop_foot, ps_loc = hq },
             PlayerState { ps_type = PT_cop_car, ps_loc = st_end st })]
        mk_hq_edges st | st_start st == hq && not (st_car_only st) =
           -- cops can take a car from hq to end of any normal street
           [(PlayerState { ps_type = PT_cop_foot, ps_loc = hq },
             PlayerState { ps_type = PT_cop_car, ps_loc = st_end st }),
           -- cops can leave a car at hq and walk to end of any normal street
            (PlayerState { ps_type = PT_cop_car, ps_loc = hq },
             PlayerState { ps_type = PT_cop_foot, ps_loc = st_end st })]
        mk_hq_edges st | st_end st == hq && not (st_car_only st) =
           -- cops can leave a car at hq and walk to start of any normal street
           [(PlayerState { ps_type = PT_cop_car, ps_loc = hq },
             PlayerState { ps_type = PT_cop_foot, ps_loc = st_start st })]
        mk_hq_edges st | otherwise = []
        stationary_hq_edges =
           -- cops can leave a car at hq and stay there on foot
           [(PlayerState { ps_type = PT_cop_car, ps_loc = hq },
             PlayerState { ps_type = PT_cop_foot, ps_loc = hq }),
           -- cops can get into a car at hq and stay there
            (PlayerState { ps_type = PT_cop_foot, ps_loc = hq },
             PlayerState { ps_type = PT_cop_car, ps_loc = hq })]
        no_op_edges = [(ps, ps) | ps <- all_states]
        all_graph_edges = [(ps_to_node start, ps_to_node end, 1)
                           | (start, end) <- concat
                                [stationary_hq_edges, no_op_edges,
                                 concatMap mk_hq_edges streets,
                                 concatMap street_to_edge streets]]
        uniq_graph_edges = nub (sort all_graph_edges)
    in  (mkGraph uniq_nodes uniq_graph_edges, ps_to_node, node_to_ps)

type LocId = Int
type LocSet = Set.Set LocId

type ImpossibleList = [LocSet]

lsSize ls = Set.size ls
lsMember x ls = Set.member x ls

lsEmpty = Set.empty
lsSingleton x = Set.singleton x

lsFromList l = Set.fromList l
lsToList ls = Set.toList ls

--lsSingleton x = Set.singleton x
lsDelete ls1 x = Set.delete x ls1

lsUnion ls1 ls2 = Set.union ls1 ls2
lsUnions lsList = Set.unions lsList
lsDifference ls1 ls2 = Set.difference ls1 ls2
lsIntersection ls1 ls2 = Set.intersection ls1 ls2

type PlayerStateLocSetMap = PlayerState -> LocSet
